/*
** tga.c -- Targa file handling routines for ShowTime
**
** Copyright (C) 1994 by Michael J. Gourlay
**
** NO WARRANTEES, EXPRESS OR IMPLIED.
*/

#include "showtime.h"
#include "file.h"
#include "image.h"
#include "cmap.h"
#include "tga.h"

/* ====================================================================== */

/* load_tga_cmap_entry: load a Targa colormap value from fio into rgb
** size tells the size of the RGB tuple
** Note that the value order in file is BGR not RGB !!!
*/
static int
load_tga_cmap_entry(FILE *fio, unsigned char rgb[], int size)
{
  switch(size) {
    case 8:
      /* gray scale */
      GET_BYTE(fio, rgb[R]);
      rgb[G] = rgb[B] = rgb[R];
      break;

    case 16: case 15:
      {
        unsigned char ip, jp;

        /* 5 bits of each red, green, and blue */
        GET_BYTE(fio, ip);
        GET_BYTE(fio, jp);

        rgb[R] = (jp & 0x7c) >> 2;
        rgb[G] = ((jp & 0x03) << 3) + ((ip & 0xe0) >> 5);
        rgb[B] = ip & 0x1f;
      }
      break;

    case 32: case 24:
      /* 8 bits of each red, green, and blue */
      GET_BYTE(fio, rgb[B]);
      GET_BYTE(fio, rgb[G]);
      GET_BYTE(fio, rgb[R]);
      if(size == 32) {
        unsigned char alpha;
        GET_BYTE(fio, alpha); /* ignore alpha */
      }
      break;

    default:
      logmesg("load_tga_cmap_entry: unknown cmap size %i\n", size);
      return(1);
      break;
  }
  return(0);
}


/* ---------------------------------------------------------------------- */

#define CURR_ROW(irow) ((tgaP->origin_bit)?(irow):(ihdrP->num_rows-(irow)-1))

#define ROW_INC \
{ \
  row_count++; \
  if(tgaP->interleave==TGA_IL_Four) { ilace_row += 4 ; } \
  else if(tgaP->interleave==TGA_IL_Two) { ilace_row += 2 ; } \
  else  { ilace_row ++ ; } \
  if(ilace_row > ihdrP->num_rows) ilace_row = ++ top_row; \
  c_row = CURR_ROW(ilace_row); \
}

/* ---------------------------------------------------------------------- */

/* load_tga_pixel: read a Targa pixel from fio and store it in img
** mapped tells whether bits are gray, coded RGB, or cmap index
**
** Full color targa formats are converted to 8-bit colormapped
** truncated RGB 332 color space, i.e. 3 red, 3 green, 2 blue bits per
** pixel.  In order for this to display correctly, the RGB332 colormap
** must be installed.  Some error diffusion is done in approximating
** the full color space to the RGB332 space, and although floyd-steinberg
** would be better, my way is much faster and neater and makes good results.
** However, in RLE encoded images, to improve speed, error diffusion
** is not performed per pixel.
**
** note: This code is optimized for speed, not readability or compactness
** so the redundancy is intentional.
**
** -- MJG 12apr94
*/
static unsigned char rgb332[256][3];
static int rgb332_allocated=FALSE;

static int
load_tga_pixel(FILE *fio, unsigned char *img, int npixels, int size, int mapped)
{
  if(!rgb332_allocated) {
    make_palette(RGB332, rgb332);
    rgb332_allocated=TRUE;
  }

  switch(size) {
    case 8:
      if(get_block(fio, img, npixels)) {
        logmesg("\nload_tga_pixel: read error\n");
        return(-1);
      }
      break;

    case 16: case 15:
      {
        int           pcount;
        int           ip, jp;
        unsigned char rb, gb, bb;
        int           ri, gi, bi;
        int           rerr=0, gerr=0, berr=0;
        int           pixel;

        for(pcount=0; pcount < npixels; pcount++) {
          GET_BYTE(fio, ip);
          GET_BYTE(fio, jp);
          if(mapped) {
            pixel = ((unsigned int) jp << 8) + ip;
          } else {
            /* Unpack color bits (5 each for red, green, blue */
            ri = (jp & 0x7c) >> 2;
            gi = ((jp & 0x03) << 3) + ((ip & 0xe0) >> 5);
            bi = ip & 0x1f;

            RGB332MAP(pixel);
          }
          img[pcount] = pixel;
        }
      }
      break;

    case 32: case 24:
      {
        int           pcount;
        int           alpha;
        unsigned char rb, gb, bb;
        int           ri, gi, bi;
        int           rerr=0, gerr=0, berr=0;
        int           pixel;

        for(pcount=0; pcount < npixels; pcount++) {
          GET_BYTE(fio, bi);
          GET_BYTE(fio, gi);
          GET_BYTE(fio, ri);
          if(size == 32) {
            GET_BYTE(fio, alpha); /* ignore alpha */
          }
          RGB332MAP(pixel);

          img[pcount] = pixel;
        }
      }
      break;

    default:
      logmesg("load_tga_pixel: unknown pixel size %i\n", size);
      return(-1);
      break;

  }
  return(0);
}

/* ---------------------------------------------------------------------- */

/* load_tga_image: Load a Targa image file from fio into sdP
*/
int
load_tga_image(image_hdr_t *ihdrP, screendata_t *sdP, FILE *fio)
{
  int           col;
  unsigned char img_byte;
  int           rle_count;      /* run-length of data */
  int           c_row;          /* current row index being loaded */
  int           rle_compressed; /* whether run is encoded or literal */
  int           blir;           /* bytes left in row */
  int           nbytes;         /* number of bytes to read */
  int           rl_encoded;     /* boolean flag */
  int           row_count;      /* total count of rows loaded */
  int           ilace_row;      /* interlaced row counter */
  int           top_row;        /* where to start over for interlaced images */
  tga_hdr_t     *tgaP;

  tgaP = ihdrP->hdr;

  if(tgaP->img_type == TGA_RLE_Map || tgaP->img_type == TGA_RLE_RGB ||
     tgaP->img_type == TGA_RLE_Mono)
  {
    rl_encoded = 1;
  } else {
    rl_encoded = 0;
  }

  row_count = ilace_row = top_row = 0;
  c_row = CURR_ROW(ilace_row);

  if(rl_encoded) {
    while(row_count < ihdrP->num_rows) {
      for(col=0 ; col < ihdrP->num_cols; ) {
        /* Read in the RLE count */
        GET_BYTE(fio, img_byte);
        if(img_byte & 0x80) {
          /* run-length encoded pixel */
          rle_count = img_byte - 127;
          rle_compressed = 1;

          /* Read the repeated byte */
          if(load_tga_pixel(fio, &img_byte, 1, tgaP->pixel_size, tgaP->mapped))
          {
            logmesg("load_tga_image: read error in rle row %i\n", c_row);
            return(EOF);
          }

        } else {
          /* stream of unencoded pixels */
          rle_count = img_byte + 1;
          rle_compressed = 0;
        }

        blir = ihdrP->num_cols - col;

        /* Put run data into image memory */
        while(rle_count) {
          if(rle_count <= blir) {
            /* finish the RLE block */
            nbytes = rle_count;
          } else {
            /* finish the row */
            nbytes = blir;
          }
          if(row_count >= ihdrP->num_rows) {
            logmesg("load_tga_image: overread image.\n");
	    col = ihdrP->num_cols;
            break;
          }
          if(rle_compressed) {
            memset(sdP->data[c_row]+col+sdP->header_size, img_byte, nbytes);
          } else {
            if(load_tga_pixel(fio, sdP->data[c_row]+col+sdP->header_size,
                 nbytes, tgaP->pixel_size, tgaP->mapped) )
            {
              logmesg("load_tga_image: read err 3 in row %i\n", c_row);
              return(EOF);
            }
          }
          if(rle_count <= blir) {
            col += rle_count;
            rle_count = 0;
          } else {
            rle_count -= blir;
            col = 0;
            blir = ihdrP->num_cols;
            ROW_INC;
          }
        } /* while */
      } /* for col */

      ROW_INC;
    } /* while row_count */

  }  else {
    /* Not run-length encoded */
    /* load pixel data one row at a time */
    while(row_count < ihdrP->num_rows) {
      if(load_tga_pixel(fio, sdP->data[c_row]+sdP->header_size, sdP->num_cols,
           tgaP->pixel_size, tgaP->mapped) )
      {
        logmesg("\nload_tga_image: read error in row %i\n", c_row);
        return(EOF);
      }
      ROW_INC;
    }
  }

  return(0);
}

/* ---------------------------------------------------------------------- */

/* load_tga_header: load a Targa image header from fio into ihdrP and sdP
**
** Note: since Targa files (aka tga files) have no magic number
** it is not a simple matter to determine whether a file is a valid
** Targa image.  Therefore, there are several consistency checks in
** this header reading routine to try to determine whether the file
** is a valid targa file.
**
** If this routine returns nonzero, then either the file is not a
** valid targa file, or we don't support this type.
**
** In the case that the file is not a Targa file, then you could
** lseek to the beginning of the file and try to read it as another
** type of image.
**
** Since there is no way to be certain of whether the error is because
** this is not a Targa at all, or if it is because the file is simply
** a corrupt or unsupported Targa, no error messages are reported
** by this routine.  Instead, a different value is returned for
** every different kind of reason why this routine rejected the
** header.  The caller routine is responsible for handling this return
** value appropriately.
*/
int
load_tga_header(image_hdr_t *ihdrP, screendata_t *sdP, FILE *fio)
{
  unsigned char flags;
  tga_hdr_t *tgaP = ihdrP->hdr;

  /* Allocate targa header */
  if((tgaP = ihdrP->hdr = (tga_hdr_t *)malloc(sizeof(tga_hdr_t)))==NULL) {
    logmesg("load_tga_header: out of memory allocating tgaP\n");
    exit(1);
  }

  ihdrP->image_type = TARGA_MAGIC;

  GET_BYTE(fio, tgaP->id_len);
  GET_BYTE(fio, tgaP->cmap_type);
  GET_BYTE(fio, tgaP->img_type);

  /* Verify that this is among the supported Targa types */
  switch(tgaP->img_type) {
    case TGA_Map:  case TGA_RLE_Map:
    case TGA_RGB:  case TGA_RLE_RGB:
    case TGA_Mono: case TGA_RLE_Mono:
      break;

    default:
      /* This is not a Targa I can deal with */
      /* (or it is not a Targa at all) */
      return(1);
      break;
  }

  /* Load rest of Targa header */
  tgaP->cmap_index = get_le_word(fio);
  tgaP->cmap_len = get_le_word(fio);
  GET_BYTE(fio, tgaP->cmap_size);
  sdP->xoffset = tgaP->x_off = get_le_word(fio);
  sdP->yoffset = tgaP->y_off = get_le_word(fio);
  ihdrP->num_cols = get_le_word(fio);
  ihdrP->num_rows = get_le_word(fio);
  GET_BYTE(fio, tgaP->pixel_size);

  GET_BYTE(fio, flags);
  tgaP->att_bits = flags & 0xf;
  tgaP->reserved = (flags & 0x10) >> 4;
  tgaP->origin_bit = (flags & 0x20) >> 5;
  tgaP->interleave = (flags & 0xc0) >> 6;

  /* Load the ID field */
  if(tgaP->id_len) {
    unsigned char id_field[256];

    if(get_block(fio, id_field, tgaP->id_len)) {
      logmesg("load_tga_header: read error in id field\n");
      return(EOF);
    }
  }

  /* Verify the validity of the colormap or pixel size */
  if(tgaP->img_type == TGA_Map || tgaP->img_type == TGA_RLE_Map
     || tgaP->img_type == TGA_CompMap || tgaP->img_type == TGA_CompMap4 )
  {
    if(tgaP->cmap_type != 1) {
      /* There was no valid colormap, but one was required */
      return(2);
    }

    tgaP->mapped = 1;

    switch(tgaP->cmap_size) {
      case 8:
      case 24: case 32:
        ihdrP->load_img = load_tga_image;
        break;

      case 15: case 16:
        ihdrP->load_img = load_tga_image;
        break;

      default:
        /* invalid colormap entry size */
        return(3);
        break;
    }
  } else {

    tgaP->mapped = 0;

    switch(tgaP->pixel_size) {
      case 8:
        ihdrP->load_img = load_tga_image;
        break;

      case 15: case 16:
      case 24: case 32:
        ihdrP->load_img = load_tga_image;
        break;

      default:
        /* invalid pixel size */
        return(4);
        break;
    }
  }

  if(tgaP->cmap_type) {
    if(tgaP->cmap_index + tgaP->cmap_len > 16384) {
      /* colormap is invalid length */
      return(5);
    }
    if(tgaP->cmap_index + tgaP->cmap_len > 256) {
      /* colormap will not fit */
      return(6);
    }
    {
      int pe;
      unsigned char cmap[256][3];
      for(pe=tgaP->cmap_index; pe < (tgaP->cmap_index + tgaP->cmap_len); pe++) {
        load_tga_cmap_entry(fio, &(cmap[pe][0]), tgaP->cmap_size);
      }

      /* Install the colormap */
      install_cmap(pe, cmap, 255);
    }
  }

  return(0);
}
